from web3 import Web3
import eth_account

class Signer:
    def __init__(self):
        pass
        

    def get_eip712_hash(self, domain_hash, struct_hash):
        """
            Returns the EIP712 hash.
            Inputs:
                - domain_hash: chain domain hash
                - struct_hash: struct hash of information to be signed
        """
        return Web3.solidity_keccak(
        [
            'bytes2',
            'bytes32',
            'bytes32'
        ],
        [
            '0x1901',
            domain_hash,
            struct_hash
        ]
    ).hex()


    def sign_hash(self, hash, private_key, append=''):
        """
            Signs the hash and returns the signature. 
        """
        result = eth_account.account.Account.sign_message(
            eth_account.messages.encode_defunct(hexstr=hash),
            private_key
        )
        return result['signature'].hex() + append



import socketio
import time
from .enumerations import MARKET_SYMBOLS, SOCKET_EVENTS
import asyncio
sio = socketio.Client()


class Sockets:
    callbacks = {}

    def __init__(self, url, timeout=10, token=None) -> None:
        self.url = url
        self.timeout = timeout
        self.token = token
        self.api_token = ""
        return

    def _establish_connection(self):
        """
            Connects to the desired url
        """
        try:
            sio.connect(self.url, wait_timeout=self.timeout,
                        transports=["websocket"])
            return True
        except:
            return False

    def set_token(self, token):
        """
            Sets default user token
            Inputs:
                - token (user auth token): firefly onboarding token.
        """
        self.token = token

    def set_api_token(self, token):
        """
            Sets default user token
            Inputs:
                - token (user auth token): firefly onboarding token.
        """
        self.api_token = token

    async def open(self):
        """
            opens socket instance connection
        """
        self.connection_established = self._establish_connection()
        if not self.connection_established:
            await self.close()
            raise (Exception("Failed to connect to Host: {}".format(self.url)))
        return

    async def close(self):
        """
            closes the socket instance connection
        """
        sio.disconnect()
        return

    @sio.on("*")
    def listener(event, data):
        """
            Listens to all events emitted by the server
        """
        try:
            if event in Sockets.callbacks.keys():
                Sockets.callbacks[event](data)
            elif "default" in Sockets.callbacks.keys():
                Sockets.callbacks["default"]({"event": event, "data": data})
            else:
                pass
        except:
            pass
        return
        
    @sio.event
    def connect():
        print("Connected To Socket Server")
        # add 10 seconds sleep to allow connection to be established before callbacks for connections are executed
        if 'connect' in Sockets.callbacks:
            # Execute the callback using asyncio.run() if available
            time.sleep(10)
            asyncio.run(Sockets.callbacks['connect']())
        

    @sio.event
    def disconnect():
        print('Disconnected From Socket Server')
        if 'disconnect' in Sockets.callbacks:
            # Execute the callback using asyncio.run() if available
            asyncio.get_running_loop().create_task(Sockets.callbacks['disconnect']())
            


    async def listen(self, event, callback):
        """
            Assigns callbacks to desired events
        """
        Sockets.callbacks[event] = callback
        return

    async def subscribe_global_updates_by_symbol(self, symbol: MARKET_SYMBOLS):
        """
            Allows user to subscribe to global updates for the desired symbol.
            Inputs:
                - symbol: market symbol of market user wants global updates for. (e.g. DOT-PERP)
        """
        try:
            if not self.connection_established:
                raise Exception(
                    "Socket connection is established, invoke socket.open()")

            resp = sio.call('SUBSCRIBE', [
                {
                    "e": SOCKET_EVENTS.GLOBAL_UPDATES_ROOM.value,
                    "p": symbol.value,
                },
            ])
            return resp["success"]
        except Exception as e:
            print("Error: ", e)
            return False

    async def unsubscribe_global_updates_by_symbol(self, symbol: MARKET_SYMBOLS):
        """
            Allows user to unsubscribe to global updates for the desired symbol.
                Inputs:
                    - symbol: market symbol of market user wants to remove global updates for. (e.g. DOT-PERP)
        """
        try:
            if not self.connection_established:
                return False

            resp = sio.call('UNSUBSCRIBE', [
                {
                    "e": SOCKET_EVENTS.GLOBAL_UPDATES_ROOM.value,
                    "p": symbol.value,
                },
            ])

            return resp["success"]
        except Exception as e:
            print(e)
            return False

    async def subscribe_user_update_by_token(self, parent_account: str = None, user_token: str = None) -> bool:
        """
            Allows user to subscribe to their account updates.
            Inputs:
                - parent_account(str): address of parent account. Only whitelisted 
                  sub-account can listen to its parent account position updates
                - token(str): auth token generated when onboarding on firefly
        """
        try:
            if not self.connection_established:
                return False

            resp = sio.call("SUBSCRIBE", [
                {
                    "e": SOCKET_EVENTS.USER_UPDATES_ROOM.value,
                    'pa': parent_account,
                    "t": self.token if user_token == None else user_token,
                    "rt": self.api_token
                },
            ])

            return resp["success"]
        except Exception as e:
            print(e)
            return False

    async def unsubscribe_user_update_by_token(self, parent_account: str = None, user_token: str = None):
        """
            Allows user to unsubscribe to their account updates.
            Inputs:
                - parent_account(str): address of parent account. Only for sub-accounts
                - token: auth token generated when onboarding on firefly
        """
        try:
            if not self.connection_established:
                return False

            resp = sio.call("UNSUBSCRIBE", [
                {
                    "e": SOCKET_EVENTS.USER_UPDATES_ROOM.value,
                    'pa': parent_account,
                    "t": self.token if user_token == None else user_token,
                    "rt": self.api_token
                },
            ])
            return resp["success"]
        except:
            return False



import json
import aiohttp
from .interfaces import *


class APIService():
    def __init__(self, url, UUID=""):
        self.server_url = url
        self.auth_token = None
        self.api_token = None
        self.uuid = UUID
        self.client = aiohttp.ClientSession()

    async def close_session(self):
        if self.client is not None:
            return await self.client.close()

    async def get(self, service_url, query={}, auth_required=False):
        """
            Makes a GET request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - query(dict): the get query.
                - auth_required(bool): indicates whether authorization is required for the call or not.
        """
        url = self._create_url(service_url)

        response = None
        if auth_required:
            response = await self.client.get(
                url,
                params=query,
                headers={
                    'Authorization': 'Bearer {}'.format(self.auth_token) if self.auth_token else '',
                    'x-api-token': self.api_token or '',
                    'x-mm-id': self.uuid or ''

                }
            )
        else:
            response = await self.client.get(url, params=query)

        try:
            if response.status != 503:  # checking for service unavailitbility
                return await response.json()
            else:
                return response
        except:
            raise Exception("Error while getting {}: {}".format(url, response))

    async def post(self, service_url, data, auth_required=False, contentType=""):
        """
            Makes a POST request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - data(dict): the data to post with POST request.
                - auth_required(bool): indicates whether authorization is required for the call or not.
        """
        url = self._create_url(service_url)
        response = None
        if auth_required:
            headers = {
                'Authorization': 'Bearer {}'.format(self.auth_token),
                'x-mm-id': self.uuid or ''
            }

            if contentType is not "":
                headers['Content-type'] = contentType

            response = await self.client.post(url=url, data=data, headers=headers)
        else:
            response = await self.client.post(url=url, data=data)

        try:
            if response.status != 503:  # checking for service unavailitbility
                return await response.json()
            else:
                return response
        except:
            raise Exception(
                "Error while posting to {}: {}".format(url, response))

    async def delete(self, service_url, data, auth_required=False):
        """
            Makes a DELETE request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - data(dict): the data to post with POST request.
                - auth_required(bool): indicates whether authorization is required for the call or not.
        """
        url = self._create_url(service_url)

        response = None
        if auth_required:
            response = await self.client.delete(
                url=url,
                data=data,
                headers={'Authorization': 'Bearer {}'.format(self.auth_token),
                 'x-mm-id': self.uuid or ''})
        else:
            response = await self.client.delete(url=url, data=data)

        try:
            return await response.json()
        except:
            raise Exception(
                "Error while posting to {}: {}".format(url, response))

    '''
        Private methods
    '''

    def _create_url(self, path):
        """
            Appends namespace to server url
        """
        return "{}{}".format(self.server_url, path)



Networks = {
  "DEV": {
    "url": "https://l2-dev.firefly.exchange/",
    "chainId": 78602,
    "apiGateway": "https://dapi-dev.firefly.exchange",
    "socketURL": "wss://dapi-dev.firefly.exchange",
    "onboardingUrl": "https://dev.firefly.exchange",
  },

  "TESTNET_ARBITRUM": {
    "url": "https://goerli-rollup.arbitrum.io/rpc",
    "chainId": 421613,
    "apiGateway": "https://dapi.api.arbitrum-staging.firefly.exchange",
    "dmsURL": "https://api.arbitrum-staging.firefly.exchange/dead-man-switch",
    "socketURL": "wss://dapi.api.arbitrum-staging.firefly.exchange",
    "webSocketURL": "wss://notifications.api.arbitrum-staging.firefly.exchange",
    "onboardingUrl": "https://testnet.firefly.exchange",
    "UUID": "uuid-default"
  },

  "MAINNET_BOBA": {
    "url": "https://bobabeam.boba.network/",
    "chainId": 1294,
    "apiGateway": "https://dapi.firefly.exchange",
    "socketURL": "wss://dapi.firefly.exchange",
    "onboardingUrl": "https://trade.firefly.exchange",
  },
  "MAINNET_ARBITRUM": {
    "url": "https://arb1.arbitrum.io/rpc/",
    "chainId": 42161,
    "apiGateway": "https://dapi.api.arbitrum-prod.firefly.exchange",
    "dmsURL": "https://api.arbitrum-prod.firefly.exchange/dead-man-switch",
    "socketURL": "wss://dapi.api.arbitrum-prod.firefly.exchange",
    "webSocketURL": "wss://notifications.api.arbitrum-prod.firefly.exchange",
    "onboardingUrl": "https://trade-arb.firefly.exchange",
    "UUID": "uuid-default"
  },
}


EIP712_DOMAIN_NAME = "IsolatedTrader"


EIP712_DOMAIN_STRING = "EIP712Domain(string name,string version,uint128 chainId,address verifyingContract)"


EIP712_ORDER_STRUCT_STRING = \
    "Order(" +  \
    "bytes8 flags," + \
    "uint128 quantity," + \
    "uint128 price," + \
    "uint128 triggerPrice," + \
    "uint128 leverage," + \
    "address maker," + \
    "uint128 expiration" + \
    ")"

ORDER_FLAGS = {
    "IS_BUY":1,
    "IS_DECREASE_ONLY": 2
}

TIME = {
  "SECONDS_IN_A_MINUTE": 60,
  "SECONDS_IN_A_DAY": 86400,
  "SECONDS_IN_A_MONTH": 2592000
}

ADDRESSES = {
  "ZERO": "0x0000000000000000000000000000000000000000",
}

SERVICE_URLS = {
  "MARKET": {
    "ORDER_BOOK": "/orderbook",
    "RECENT_TRADE": "/recentTrades",
    "CANDLE_STICK_DATA": "/candlestickData",
    "EXCHANGE_INFO": "/exchangeInfo",
    "MARKET_DATA": "/marketData",
    "META": "/meta",
    "STATUS": "/status",
    "SYMBOLS": "/marketData/symbols",
    "CONTRACT_ADDRESSES": "/marketData/contractAddresses",
    "TICKER": "/ticker",
    "MASTER_INFO": "/masterInfo",
    "FUNDING_RATE":"/fundingRate"
  },
  "USER": {
    "USER_POSITIONS": "/userPosition",
    "USER_TRADES": "/userTrades",
    "ORDERS": "/orders",
    "GENERATE_READONLY_TOKEN": "/generateReadOnlyToken",
    "ACCOUNT": "/account",
    "USER_TRANSACTION_HISTORY": "/userTransactionHistory",
    "AUTHORIZE": "/authorize",
    "ADJUST_LEVERAGE": "/account/adjustLeverage",
    "FUND_GAS": "/account/fundGas",
    "TRANSFER_HISTORY": "/userTransferHistory",
    "FUNDING_HISTORY": "/userFundingHistory",
    "CANCEL_ON_DISCONNECT": "/dms-countdown"
  },
  "ORDERS": {
    "ORDERS": "/orders",
    "ORDERS_HASH": "/orders/hash",
  },
  "GROWTH": {
    "REFERRER_INFO": "/growth/getReferrerInfo",
    "CAMPAIGN_DETAILS": "/growth/campaignDetails",
    "CAMPAIGN_REWARDS": "/growth/campaignRewards",
    "AFFILIATE_PAYOUTS": "/growth/affiliate/payouts",
    "AFFILIATE_REFEREE_DETAILS": "/growth/affiliate/refereeDetails",
    "AFFILIATE_REFEREES_COUNT": "/growth/affiliate/refereesCount",
    "USER_REWARDS_HISTORY": "/growth/userRewards/history",
    "USER_REWARDS_SUMMARY": "/growth/userRewards/summary",
    "REWARDS_OVERVIEW": "/growth/tradeAndEarn/rewardsOverview",
    "REWARDS_DETAILS": "/growth/tradeAndEarn/rewardsDetail",
    "TOTAL_HISTORICAL_TRADING_REWARDS": "/growth/tradeAndEarn/totalHistoricalTradingRewards",
    "MAKER_REWARDS_SUMMARY": "/growth/marketMaker/maker-rewards-summary",
    "MAKER_REWARDS_DETAILS": "/growth/marketMaker/maker-rewards-detail",
    "MAKER_WHITELIST_STATUS": "/growth/marketMaker/whitelist-status",
    "GENERATE_CODE": "/growth/generateCode",
    "LINK_REFERRED_USER": "/growth/linkReferredUser"
  },
}

EIP712_CANCEL_ORDER_STRUCT_STRING ="CancelLimitOrder(string action,bytes32[] orderHashes)"

EIP712_ONBOARDING_ACTION_STRUCT_STRING = \
    'firefly(' + \
    'string action,' + \
    'string onlySignOn' + \
    ')'

EIP712_DOMAIN_STRING_NO_CONTRACT = \
    "EIP712Domain(" + \
    "string name," + \
    "string version," + \
    "uint128 chainId" + \
    ")"